from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine, select,insert
from whitebox_workflows import Vector
from ...vector_extension import VectorExtension
from .reach import Reach
from .reach_parameter import ReachParameter
from .field_info import FieldInfo
from .farm_info import FarmInfo
from .subbasin_info import SubbasinInfo
from .field_subbasin import FieldSubbasin
from .field_farm import FieldFarm
from .farm_subbasin import FarmSubbasin
from .subbasin_multiplier import SubbasinMultiplier
from ...delineation.structure import Structure
from ...names import Names
from ...bmp.bmp_areal_manure_feedlot import ArealBMPManureFeedlot
from ...bmp.bmp_areal_manure_storage import ArealBMPManureStorage
from ...bmp.bmp_reach_manure_catch_basin import ReachBMPManureCatchBasin
from ...bmp.bmp_reach_reservoir import ReachBMPReservoir
from ...bmp.bmp_reach_wetland import ReachBMPWetland
from ...bmp.bmp_type import BMPType
from ...bmp.bmp_reach import ReachBMP
from ...bmp.bmp import BMP
from ...bmp.bmp_structure_dugout import StructureBMPDugout
from ...bmp.bmp_structure_wascob import StructureBMPWascob
from ...bmp.bmp_offsite_wintering import BMPOffsiteWintering
from ...bmp.crop_rotation.crop_rotation import CropRotation
from .bmp_01_point_source import PointSource
from .bmp_02_flow_diversion import FlowDiversion
from .bmp_03_reservoir import Reservoir
from .bmp_05_riparian_buffer import RiparianBuffer
# from .bmp_06_grass_waterway import GrassWaterWay
# from .bmp_07_vegetation_filter_strip import FilterStrip
from .bmp_09_isolated_wetland import Wetland
# from .bmp_12_crop_management import CropManagement, CropParameter
# from .bmp_14_tillage_management import TillageManagement, TillageParameter
# from .bmp_15_fertilizer_management import FertlizerManagement, FertilizerParameter
# from .bmp_16_grazing_management import GrazingManagement
# from .bmp_18_irrigation_management import irrigation_management, irrigation_parameter
# from .bmp_21_manure_incorporation_within_48h import ManureIncorporationWithin48hManagemet
# from .bmp_22_manure_application_setback import ManureApplicationSetbackManagement
# from .bmp_23_no_application_on_snow import ManureNoApplicaitonOnSnowManagement
# from .bmp_24_spring_application_rather_than_fall_application import ManureSpringApplicationRatherThanFallApplicationManagement
# from .bmp_25_applicaiton_based_on_soil_nitrogen_limit import ManureApplicationBasedOnSoliNitrogenLimitManagement
# from .bmp_26_applicaiton_based_on_soil_phosphorous_limit import ManureApplicationBasedOnPhosphorousLimitManagement
from .bmp_27_manure_storage import ManureStorageParameter
from .bmp_28_manure_catch_basin import ManureCatchBasinParameter
from .bmp_29_manure_feedlot import ManureFeedlot, ManureFeedlotManagement
# from .bmp_30_marginal_crop_management import MarginalCropManagement
# from .bmp_31_marginal_fertilizer_management import MarginalFertilizerManagement
# from .bmp_32_marginal_tillage_management import MarginalTillageManagement
# from .bmp_33_wintering_site_management import WinteringSiteManagement, WinteringSiteParameter
# from .bmp_34_pasture_crop_management import PastureCropManagement
# from .bmp_35_pasture_fertilizer_management import PastureFertilizerManagement
# from .bmp_36_pasture_tillage_management import Pasture_tillage_management
# from .bmp_37_pasture_grazing_management import PastureGrazingManagement, PastureGrazingParameter
# from .bmp_38_dugout import DugoutParameter
from .bmp_39_offsite_watering import OffsiteWateringParameter
from .bmp_40_managed_access_including_fencing import ManagedAccessIncludingFencingParameter
from .bmp_41_wascob import Wascob
#from .bmp_42_water_use import 
import os


from ..database_base import DatabaseBase, logger
from .bmp_table import BMPTable
from ...raster_extension import RasterExtension
from whitebox_workflows import WbEnvironment, Raster
import pandas as pd
import math
import numpy as np
from ...outputs import Outputs

      

class BMPDatabase(DatabaseBase):
    default_tables = [Names.bmp_table_name_bmp_index,
                      Names.bmp_table_name_crop_parameter,
                      Names.bmp_table_name_crop_remove_parameter,
                      Names.bmp_table_name_fertilizer_parameter,
                      Names.bmp_table_name_tillage_parameter,
                      Names.bmp_table_name_livestock_parameter,                      
                      Names.bmp_table_name_ls_parameter]

    COL_NAME_AREA_HA = "Area_Ha"

    def __init__(self, database_file):
        super().__init__(database_file)
        self.wbe = WbEnvironment()   
    
    def create_database_structure(self):
        #remove the tables we want to update and then create them again
        logger.info("Creating table structure in bmp database ...")
        BMPTable.metadata.drop_all(self.engine)
        BMPTable.metadata.create_all(self.engine)

        #default tables
        logger.info("Loading default tables to bmp database ...")
        for table in BMPDatabase.default_tables:
            logger.info(table)
            self.populate_defaults(table) 

    def create_bmp_tables(self, outputs:Outputs, 
                          reservoir_flow_routing:str, reservoir_flow_data_folder:str):
        """
        create all the parameter tables in bmp database
        """        

        self.__create_bmp_manure_feedlot(outputs)
        self.__create_bmp_manure_catchbasin(outputs)
        self.__create_bmp_manure_storage(outputs)

        self.__create_bmp_reservoir(outputs,reservoir_flow_routing, reservoir_flow_data_folder)
        self.__create_bmp_wetland(outputs)
        self.__create_bmp_dugout(outputs)
        self.__create_bmp_wascob(outputs)
        self.__create_bmp_riparian_buffer(outputs)

        self.__create_bmp_offsite_wintering(outputs)

        #create reach_bmp table
        #all reach bmps should be added before this
        self.__create_reach_parameter(outputs)
        self.__create_bmp_reach_bmp(outputs)

        #create bmp_scenarios table
        self.__create_bmp_scenarios(outputs)

#region spatial relationship

    def create_spatial_relationship_tables(self,outputs:Outputs):
        logger.info("Creating field_info ...")
        field_area_df = RasterExtension.get_category_area_ha_dataframe(outputs.field_raster, BMPDatabase.COL_NAME_AREA_HA)
        self.save_table(Names.bmp_table_name_field_info, field_area_df, None, True)
        field_area_df.columns = ["FieldArea"]

        logger.info("Creating farm_info ...")
        farm_area_df = RasterExtension.get_category_area_ha_dataframe(outputs.farm_raster, BMPDatabase.COL_NAME_AREA_HA)
        self.save_table(Names.bmp_table_name_farm_info, farm_area_df, None, True)
        farm_area_df.columns = ["FarmArea"]

        logger.info("Creating subbasin_info ...")
        subbasin_area_df = self.__create_subbasin_info(outputs.subbasin_raster, 
                                                       outputs.flow_acc_raster, 
                                                       outputs.potential_ruoff_coefficient_raster, 
                                                       outputs.depression_storage_capacity_raster, 
                                                       outputs.cn2_raster)
        self.save_table(Names.bmp_table_name_subbasin_info, subbasin_area_df, None, True)
        subbasin_ids = subbasin_area_df.index.to_list()

        subbasin_area_df = subbasin_area_df[BMPDatabase.COL_NAME_AREA_HA].to_frame()
        subbasin_area_df.columns = ["SubbasinArea"]

        logger.info("Creating field_subbasin ...")
        self.__create_overlay(outputs.field_raster, outputs.subbasin_raster, 
                              "Field", "Subbasin", field_area_df, subbasin_area_df, Names.bmp_table_name_field_subbasin)

        logger.info("Creating farm_subbasin ...")
        self.__create_overlay(outputs.farm_raster, outputs.subbasin_raster, 
                              "Farm", "Subbasin", farm_area_df, subbasin_area_df, Names.bmp_table_name_farm_subbasin)

        logger.info("Creating field_farm ...")
        self.__create_overlay(outputs.field_raster, outputs.farm_raster, 
                              "Field", "Farm", field_area_df, farm_area_df, Names.bmp_table_name_field_farm) 
        
        self.__create_subbasin_multiplier(subbasin_ids)        

    def __create_subbasin_info(self, 
                                subbasin_raster:Raster,
                                flow_acc_raster:Raster, 
                                potential_runoff_coefficient_raster:Raster, 
                                depression_storage_capacity_raster:Raster, 
                                cn2_raster:Raster):
        
        id_area_df = RasterExtension.get_category_area_ha_dataframe(subbasin_raster, BMPDatabase.COL_NAME_AREA_HA)
        id_flow_acc_max_df = RasterExtension.get_zonal_statistics(flow_acc_raster, subbasin_raster, "max", "FlowAcc_Max")
        id_prc_avg_df = RasterExtension.get_zonal_statistics(potential_runoff_coefficient_raster, subbasin_raster, "mean", "PRC_Avg")
        id_dsc_avg_df = RasterExtension.get_zonal_statistics(depression_storage_capacity_raster, subbasin_raster,"mean", "DSC_Avg")
        id_cn2_df = RasterExtension.get_zonal_statistics(cn2_raster, subbasin_raster,"mean", "CN2_Avg")
        
        #merge all info
        mergered_df = id_area_df.merge(id_flow_acc_max_df, left_index=True, right_index=True, how="inner")
        mergered_df = mergered_df.merge(id_prc_avg_df, left_index=True, right_index=True, how="inner")
        mergered_df = mergered_df.merge(id_dsc_avg_df, left_index=True, right_index=True, how="inner")
        mergered_df = mergered_df.merge(id_cn2_df, left_index=True, right_index=True, how="inner")

        return mergered_df

    def __create_overlay(self, spatial1_ras:Raster, spatial2_ras:Raster, name1:str, name2:str, raster1_area_df:pd.DataFrame, raster2_area_df:pd.DataFrame, table_name:str):
        overlay_area_df = self.__get_overlay_area(spatial1_ras, spatial2_ras, name1, name2)
        
        merged_df = overlay_area_df.merge(raster1_area_df, left_on= name1, right_index=True, how="left")
        merged_df = merged_df.merge(raster2_area_df, left_on= name2, right_index=True, how="left")

        #calculate the ratio
        merged_df["ID"] = merged_df.index
        merged_df[f"To{name1}"] = merged_df[BMPDatabase.COL_NAME_AREA_HA] / merged_df[f"{name1}Area"]
        merged_df[f"To{name2}"] = merged_df[BMPDatabase.COL_NAME_AREA_HA] / merged_df[f"{name2}Area"]

        merged_df = merged_df[["ID", name1, name2, BMPDatabase.COL_NAME_AREA_HA, f"To{name1}", f"To{name2}"]]     
        self.save_table(table_name, merged_df)   
    
    def __get_overlay_area(self, spatial1_ras:Raster, spatial2_ras:Raster, name1:str, name2:str)->pd.DataFrame:
        merged_raster, raster1_max = RasterExtension.get_overlay_raster(spatial1_ras, spatial2_ras)
        df = RasterExtension.get_category_area_ha_dataframe(merged_raster,BMPDatabase.COL_NAME_AREA_HA)
        df[name1] = df.index % raster1_max
        df[name2] = (df.index - df[name1]) / raster1_max

        return df
    
    def __create_subbasin_multiplier(self, subbasin_ids:list)->None:
        """populate default subbasin-muliplier"""
        logger.info("Creating subbasin_multiplier ... ")
        Session = sessionmaker(bind=self.engine)
        with Session() as session:
            session.add_all([SubbasinMultiplier(id) for id in subbasin_ids])
            session.commit()

    def __create_reach_parameter(self, outputs:Outputs):
        """crete reach parameters"""
        logger.info("Creating reach_parameter table ... ")
        self.save_table(Names.bmp_table_name_reach_parameter, outputs.reach_parameter_df)

#endregion
   
#region Structure BMPs
    
    def __create_bmp_dugout(self, outputs:Outputs):
        if "dugout" not in outputs.structures:
            return
        
        logger.info("Creating dugout ... ")
        dugout = StructureBMPDugout(outputs.inputs.dugout_boundary_vector, outputs.subbasin_raster, outputs.structures["dugout"])
        self.save_table(Names.bmp_table_name_dugout, dugout.dugout_df)

    def __create_bmp_wascob(self, outputs:Outputs):
        if "wascob" not in outputs.structures:
            return
        
        logger.info("Creating wascob ... ")
        wascob = StructureBMPWascob(outputs.inputs.dugout_boundary_vector, outputs.subbasin_raster, outputs.structures["wasob"])
        self.save_table(Names.bmp_table_name_wascob, wascob.wascob_df)

    def __create_bmp_riparian_buffer(self, outputs:Outputs):
        df = outputs.riparian_buffer_parameter_df
        if df is not None:
            self.save_table(Names.bmp_table_name_riparian_buffer, df, RiparianBuffer.column_types())

#endregion

#region Manure Management BMPs

    def __create_bmp_manure_feedlot(self, outputs:Outputs):
        """Create parameter and management table for manure feedlot"""

        if outputs.inputs.feedlot_boundary_vector is None:
            return
        
        #get the structure and creat feedlot
        feedlots = ArealBMPManureFeedlot(outputs.inputs.feedlot_boundary_vector,
                                         outputs.flow_direction_raster,
                                         outputs.subbasin_raster, 
                                         outputs.reach_raster, 
                                         outputs.dem_clipped_raster_for_model)

        #parameter table        
        Session = sessionmaker(bind=self.engine)
        with Session() as session:
            session.add_all(feedlots.parameters)
            session.commit()

        #management table
        self.save_table(Names.bmp_table_name_manure_feed_lot_management, feedlots.default_management_df, ManureFeedlotManagement.column_types())

    def __create_bmp_manure_catchbasin(self, outputs:Outputs):
        """Create parmater table for manure catch basin"""

        if outputs.inputs.catchbasin_vector is None:
            return {}
        
        logger.info("Creating manure catch basin parameter table ... ")
        catchbasin = ReachBMPManureCatchBasin(outputs.inputs.catchbasin_vector,
                                outputs.subbasin_raster,
                                outputs.reach_parameter_df)
        
        Session = sessionmaker(bind=self.engine)
        with Session() as session:
            session.add_all(catchbasin.manure_catch_basin_parameters)
            session.commit()         

    def __create_bmp_manure_storage(self,outputs:Outputs):
        """Create parmater and management table for manure storage"""

        if outputs.inputs.manure_storage_boundary_vector is None:
            return
        
        logger.info("Creating manure storage parameter table ... ")
        storage = ArealBMPManureStorage(outputs.inputs.manure_storage_boundary_vector,
                                outputs.flow_direction_raster,
                                outputs.subbasin_raster,
                                outputs.reach_raster,
                                outputs.dem_clipped_raster)
        
        #parameter table
        Session = sessionmaker(bind=self.engine)
        with Session() as session:
            session.add_all(storage.manure_storage_parameters)
            session.commit() 

#endregion

#region Reach BMPs

    def __create_bmp_reach_bmp(self, outputs:Outputs):
        """create reach_bmp table"""

        logger.info("Creating reach_bmp table ... ")

        reach_bmps = {} 

        #wetland      
        if "wetland" in outputs.structures:
            reach_bmps[BMPType.BMP_TYPE_WETLAND] = ReachBMPWetland(outputs.structures["wetland"].boundary_processed_vector, outputs.subbasin_raster).subbasins
        
        #reservoir
        if outputs.inputs.reservoir_vector is not None:
            reach_bmps[BMPType.BMP_TYPE_RESERVOIR] = ReachBMPReservoir(outputs.inputs.reservoir_vector, outputs.subbasin_raster).subbasins
        
        #catch basin
        reach_bmps[BMPType.BMP_TYPE_MANURE_CATCHBASIN] = ReachBMPManureCatchBasin(outputs.inputs.catchbasin_vector,
                                outputs.subbasin_raster,
                                outputs.reach_parameter_df).subbasins

        if len(reach_bmps) <= 0:
            logger.info("There is no reach bmp, existing ... ")
            return

        #get all subbasins
        subbasin_ids = range(1, int(RasterExtension.get_max_value(outputs.subbasin_raster)) + 1)

        #save the table
        df = ReachBMP.create_reach_bmp_df(subbasin_ids, reach_bmps)
        self.save_table(Names.bmp_table_name_reach_bmp, df)

    def __create_bmp_wetland(self, outputs:Outputs):
        if "wetland" not in outputs.structures:
            return
        
        logger.info("Creating wetland ... ")
        Session = sessionmaker(bind=self.engine)
        with Session() as session:
            session.add_all([Wetland(struc_attribute) for struc_attribute in outputs.structures["wetland"].attributes.values()])
            session.commit()

    def __create_bmp_reservoir(self, outputs:Outputs, 
                           flow_method:str, 
                           flow_data_folder:str):
        """
        populate reservoir parameter table
        """

        if outputs.inputs.reservoir_vector is None:
            return {}
        
        logger.info("Creating reservoir parameter table ... ")
        ReachBMPReservoir.validate(outputs.inputs.reservoir_vector, flow_method, flow_data_folder)
        reservoir = ReachBMPReservoir(outputs.inputs.reservoir_vector, outputs.subbasin_raster)

        for res in reservoir.reservoirs:
            res.METHOD = flow_method
            logger.info(f"Importing {res.FILE}.csv ...")
            file_path = os.path.join(flow_data_folder, f"{res.FILE}.csv")
            if os.path.exists(file_path):
                self.save_table(res.FILE, pd.read_csv(file_path))

        Session = sessionmaker(bind=self.engine)
        with Session() as session:
            session.add_all(reservoir.reservoirs)
            session.commit()  

        return reservoir.subbasins

#endregion

    def __create_bmp_offsite_wintering(self, outputs:Outputs):
        if outputs.inputs.offsite_watering_vector is None:
            return
        
        logger.info("Creating offsite wintering parameter table ... ")
        bmp = BMPOffsiteWintering(outputs.inputs.offsite_watering_vector, outputs.subbasin_raster)

        Session = sessionmaker(bind=self.engine)
        with Session() as session:
            session.add_all(bmp.offsite_wintering_parameters)
            session.commit()      
  
    def __create_bmp_scenarios(self, outputs:Outputs, include_crop_rotation:bool = False):
        """create bmp scenarios table"""
        
        logger.info("Creating bmp_secenarios table ... ")

        bmp_types = outputs.inputs.bmp_types
        if include_crop_rotation:
            bmp_types.append(BMPType.BMP_TYPE_CROP)
            bmp_types.append(BMPType.BMP_TYPE_FERTILIZER)
            bmp_types.append(BMPType.BMP_TYPE_TILLAGE)
            bmp_types.append(BMPType.BMP_TYPE_GRAZING)

        df = BMP.generate_bmp_scenarios_df(bmp_types)
        self.save_table(Names.bmp_table_name_scenarios, df)

#region Crop Rotation

    def update_crop_rotation_AAFC_crop_inventory(self, crop_inventory_folder:str, first_year:int, last_year:int, outputs:Outputs):        
        logger.info("Updating crop rotation based on AAFC Crop Inventory ... ")

        rotation = CropRotation(outputs.field_clipped_vector, outputs.subbasin_raster,crop_inventory_folder, first_year, last_year)
        
        logger.info("Crop Management ... ")
        self.save_table(Names.bmp_table_name_crop_management, rotation.crop_management_df)

        logger.info("Fertilizer Management ... ")
        self.save_table(Names.bmp_table_name_fertilizer_management, rotation.fertilizer_management_df)

        logger.info("Tillage Management ... ")
        self.save_table(Names.bmp_table_name_tillage_management, rotation.tillage_management_df)

        logger.info("Grazing Management ... ")
        self.save_table(Names.bmp_table_name_grazing_management, rotation.grazing_management_df)

        logger.info("Grazing reach deposit ... ")
        self.__create_grazing_reach_deposit(first_year)

        logger.info("Updating scenarios ...")
        self.__create_bmp_scenarios(outputs, True)

    def __create_grazing_reach_deposit(self, start_year:int):
        """
        Create GRAMG_ReachDeposit table based on GRAMG_management, Field_Info, Livestock_parameter and Fertilizer_parameter.

        It pre-calculate the manure deposit at each reach on each day as a way to speed the engine calculation. Could and should be done in the engine. 

        The table name GRAMG_ReachDeposit and GRAMG_management is fixed in the engine so it couldn't be renamed, which should be avoided. 

        This replace the GenerateGrazingReachDeposit function in Whitebox-based interface. 
        
        """
        
        grazing_management_df = self.read_table("GRAMG_management")
        if grazing_management_df is None:
            return 

        field_info_df = self.read_table(Names.bmp_table_name_field_info)
        livestock_parameter = self.read_table(Names.bmp_table_name_livestock_parameter)
        fertilizer_parameter = self.read_table(Names.bmp_table_name_fertilizer_parameter)

        #inner join grazing_management, field_info and livestock_parameter
        reach_deposit_df = grazing_management_df[grazing_management_df["Source"] == 1]
        reach_deposit_df = reach_deposit_df[reach_deposit_df["Access"] != 2]
        reach_deposit_df = reach_deposit_df.merge(field_info_df, left_on="Location", right_on= "id", how="inner")
        reach_deposit_df = reach_deposit_df.merge(livestock_parameter, left_on="Ani_ID", right_on="ID", how="inner")
        reach_deposit_df = reach_deposit_df.merge(fertilizer_parameter, left_on="Man_ID", right_on="IFNUM", how="inner")

        #calcualte manure and drinkwater
        reach_deposit_df["Manure"] = reach_deposit_df["GR_Density"] * reach_deposit_df["Area_Ha"] * reach_deposit_df["Ani_Weight"] * reach_deposit_df["Ani_adult"] / 100 / 1000 * reach_deposit_df["Man_Day"] * reach_deposit_df["DayFra"] * reach_deposit_df["Drinking_time"]
        reach_deposit_df["DrinkWater"] = reach_deposit_df["GR_Density"] * reach_deposit_df["Area_Ha"] * reach_deposit_df["Ani_Weight"] * reach_deposit_df["Ani_adult"] / 100 * reach_deposit_df["Water_Drink"] / 100 / 1000 

        #further calculation and rename
        reach_deposit_df["Year"] = reach_deposit_df["Year"] + start_year - 1
        reach_deposit_df["Month"] = reach_deposit_df["GraMon"]
        reach_deposit_df["Day"] = reach_deposit_df["GraDay"]
        reach_deposit_df["StartDate"] = pd.to_datetime(reach_deposit_df[["Year","Month","Day"]])
        reach_deposit_df["ReachId"] = reach_deposit_df["SourceID"]
        reach_deposit_df["TSS"] = reach_deposit_df["Man_TSS_Fra"] / 100 * reach_deposit_df["Manure"]
        reach_deposit_df["NO3"] = reach_deposit_df["FMINN"]  * reach_deposit_df["Manure"]
        reach_deposit_df["NH4"] = 0
        reach_deposit_df["OrgN"] = reach_deposit_df["FORGN"] * reach_deposit_df["Manure"]
        reach_deposit_df["SolP"] = reach_deposit_df["FMINP"] * reach_deposit_df["Manure"]
        reach_deposit_df["OrgP"] = reach_deposit_df["FORGP"] * reach_deposit_df["Manure"]

        #just keep the useful columns
        reach_deposit_df = reach_deposit_df[["ReachId","StartDate","Days","Manure","TSS","NO3","NH4","OrgN","SolP","OrgP","DrinkWater","BankK_Change"]]

        #repeat per the days column
        expanded_reach_deposit_dfs = []
        for index in range(len(reach_deposit_df)):
            #repeat the records based on the number of days and set the date
            repeated_df = pd.DataFrame(np.repeat(reach_deposit_df.iloc[[index]].values,reach_deposit_df.iloc[index].Days,axis=0),columns=reach_deposit_df.columns).reset_index()
            temp = repeated_df["index"].apply(lambda x: pd.Timedelta(x, unit='D'))
            repeated_df["Date"] = repeated_df["StartDate"] + temp
            expanded_reach_deposit_dfs.append(repeated_df)

        #concat
        reach_deposit_df = pd.concat(expanded_reach_deposit_dfs)

        #aggregate for each and date
        reach_deposit_df = pd.pivot_table(reach_deposit_df, 
                        index = ["ReachId","Date"], 
                        values =["Manure","TSS","NO3","NH4","OrgN","SolP","OrgP","DrinkWater","BankK_Change"],
                        aggfunc={"Manure":"sum",
                                "TSS":"sum",
                                "NO3":"sum",
                                "NH4":"sum",
                                "OrgN":"sum",
                                "SolP":"sum",
                                "OrgP":"sum",
                                "DrinkWater":"sum",
                                "BankK_Change":"min"})

        reach_deposit_df.reset_index(inplace=True)

        #get year, month and day
        reach_deposit_df["Year"] = reach_deposit_df["Date"].dt.year - start_year + 1
        reach_deposit_df["Month"] = reach_deposit_df["Date"].dt.month
        reach_deposit_df["Day"] = reach_deposit_df["Date"].dt.day

        #get required columns
        reach_deposit_df = reach_deposit_df[["ReachId","Year","Month","Day","Manure","TSS","NO3","NH4","OrgN","SolP","OrgP","DrinkWater","BankK_Change"]]

        #save to database
        self.save_table("GRAMG_ReachDeposit",reach_deposit_df)

#endregion




from typing import Any
from sqlalchemy import Column, Integer, REAL
from .bmp_table import BMPTable
from ...names import Names

class GrassWaterWay(BMPTable):
    """Parameter Table for BMP: Grass waterway (6)"""
    __tablename__ = Names.bmp_table_name_grass_waterway
    """Riparian buffer drainage part id used for IMWEBs model calculation, one RIBUF may contain several drainage parts"""
    ID = Column(Integer, primary_key=True)
    """Year of operation"""
    Year = Column(Integer)
    """Subbasin ID"""
    Subbasin = Column(Integer)
    """GWW length (m)"""
    Length = Column(REAL)
    """GWW bankfull width (m), default is 10 m"""
    BFwidth = Column(REAL)
    """GWW bankfull depth (m), default is 0.5 m"""
    BFdepth = Column(REAL)
    """GWW channel side slope, default is 0.25 (1:4)"""
    SideSlope = Column(REAL)
    """GWW channel slope (%)"""
    Slope_perc = Column(REAL)
    """GWW roughness coefficient, default is 0.25"""
    Manning = Column(REAL)
    """GWW bankfull velocity (m/s)"""
    BFvelocity = Column(REAL)
    """GWW bottom conductivity (mm/hr), default is 10"""
    Conductivity = Column(REAL)
    """GWW erodibility factor, default is 0"""
    Erodibility = Column(REAL)
    """GWW cover factor, default is 0"""
    Cover = Column(REAL)

    def __init__(self):
        self.ID = 0
        self.Year = 0
        self.Subbasin = 0
        self.Length = 0
        self.BFwidth = 10
        self.BFdepth = 0.5
        self.SideSlope = 0.25
        self.Slope_perc = 0
        self.Manning = 0.25
        self.BFvelocity = 0
        self.Conductivity = 10
        self.Erodibility = 0
        self.Cover = 0




from typing import Any
from sqlalchemy import Column, Integer, Float, TEXT, REAL
from .bmp_table import BMPTable
from .bmp_16_grazing_management import GrazingManagement
from ...names import Names

class WinteringSiteParameter(BMPTable):
    """Parameter Table for Wintering Site Management"""

    __tablename__ = Names.bmp_table_name_wintering_site_parameter

    """Wintering site ID"""
    ID = Column(Integer, primary_key=True)

    """Producer ID"""
    ProducerID = Column(Integer)

    """Wintering site area (ha)"""
    Area_Ha = Column(Float)

    """Slope of the Wintering site area (%)"""
    Slope = Column(Float)

    """Distance to creek (m)"""
    Distance_m = Column(Float)

class WinteringSiteManagement(GrazingManagement):
    """Distribution Table for BMP: Wintering site management (33)"""
    def __init__(self):
        super().__init__()


# class manure_and_nutrient_management(BMPTable):
#     __tablename__ = 'manure_and_nutrient_management'
#     Scenario = Column(Integer)
#     Location = Column(Integer)
#     ID = Column(Integer)
#     Level = Column(TEXT)
#     BMP_ID = Column(Integer)
#     Parameter = Column(REAL)
#     Value = Column(REAL)

# class manure_and_nutrient_parameter(BMPTable):
#     __tablename__ = 'manure_and_nutrient_parameter'
#     BMP_ID = Column(Integer, primary_key=True)
#     Code = Column(TEXT)
#     Description = Column(TEXT)
#     Parameter = Column(TEXT)
#     Value = Column(REAL)
#     Value_TEXT = Column(TEXT)








from ..database_base import DatabaseBase
import pandas as pd
import sqlite3
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine, select
from .stations import Stations

class HydroClimateDatabase(DatabaseBase):
    """
    Hydroclimate database
    """
    def __init__(self, database_file):
        super().__init__(database_file)    

        self.__station_coordinates = {}
        self.__data_type_station_ids_dictionary = {}

    @property
    def station_coordinates(self)->dict:
        """Coordinates of all stations. Used for interpolation"""
        if len(self.__station_coordinates) <= 0:
            Session = sessionmaker(bind=self.engine)
            with Session() as session:
                select_stmt = select(Stations)
                for row in session.scalars(select_stmt):
                    self.__station_coordinates[row.ID] = (row.XPR, row.YPR)

        return self.__station_coordinates
    
    @property
    def data_type_station_ids_dictionary(self):
        """The station data type and number of stations"""
        if len(self.__data_type_station_ids_dictionary) <= 0:
            for data_type, table_appendix in HydroClimateDatabase.station_types.items():
                table_name = f"STATION_DATA_{table_appendix}"
                station_id_df = self.read_distinct_list(table_name, "STATION")
                if len(station_id_df) > 0:
                    self.__data_type_station_ids_dictionary[data_type] = station_id_df["STATION"].to_list()

        return self.__data_type_station_ids_dictionary
    
    @property
    def data_start_date(self)->pd.Timestamp:
        """The start date considering all data type"""
        start_date = pd.Timestamp('1900-01-01')
        for _, table_appendix in HydroClimateDatabase.station_types.items():
            table_name = f"STATION_DATA_{table_appendix}"
            start_date_df = pd.read_sql(f"select min(datetime(date)) start_date from {table_name}", self.engine)
            if len(start_date_df) > 0:
                data_type_start_date = pd.to_datetime(start_date_df['start_date'].to_list()[0])
                if data_type_start_date > start_date:
                    start_date = data_type_start_date

        return start_date

    @property
    def data_end_date(self)->pd.Timestamp:
        """The end date considering all data type"""
        end_date = pd.Timestamp('2100-01-01')
        for _, table_appendix in HydroClimateDatabase.station_types.items():
            table_name = f"STATION_DATA_{table_appendix}"
            end_date_df = pd.read_sql(f"select max(datetime(date)) end_date from {table_name}", self.engine)
            if len(end_date_df) > 0:
                data_type_end_date = pd.to_datetime(end_date_df['end_date'].to_list()[0])
                if data_type_end_date < end_date:
                    end_date = data_type_end_date

        return end_date


#region Migration

    sql_change_stations_table_name = """
        ALTER TABLE stations
        RENAME TO stations_old;
    """

    sql_create_stations_table = """
        CREATE TABLE IF NOT EXISTS STATIONS (
            ID INTEGER PRIMARY KEY,
            NAME TEXT NOT NULL,
            XPR  REAL,
            YPR  REAL,
            LAT  REAL,
            LONG REAL,
            ELEVATION REAL,
            AREA REAL
        );
        """

    sql_create_station_data_table = """
        CREATE TABLE IF NOT EXISTS STATION_DATA_ (
            ID INTEGER PRIMARY KEY,
            STATION INTEGER NOT NULL,
            DATE  TEXT NOT NULL,
            VALUE REAL NOT NULL,
            FOREIGN KEY (STATION) REFERENCES STATIONS (ID)
        );
        """

    sql_create_station_data_table_index = """
        CREATE INDEX IDX_STATION_DATA_ ON STATION_DATA_ (STATION);
    """

    station_data_tables = [
    "STATION_DATA_PCP",
    "STATION_DATA_TMX",
    "STATION_DATA_TMN",
    "STATION_DATA_SLR",
    "STATION_DATA_HMD",
    "STATION_DATA_WDIR",
    "STATION_DATA_WSPD"
    ]

    station_types = {
    "P":"PCP",
    "TMAX":"TMX",
    "TMIN":"TMN",
    "RM":"HMD",
    "SR":"SLR",
    "WS":"WSPD",
    "WD":"WDIR"
    }


    @staticmethod
    def run_sql(conn, create_table_sql):
        try:
            c = conn.cursor()
            c.execute(create_table_sql)
        except Exception as e:
            print(e)

    @staticmethod
    def populate_stations(conn):
        #populate the new stations table
        query = "SELECT DISTINCT NAME,XPR,YPR,LAT,LONG,ELEVATION,AREA FROM STATIONS_OLD"
        df_stations = pd.read_sql_query(query,conn)
        df_stations["ID"] = df_stations.index + 1
        df_stations.to_sql(f"STATIONS",con = conn, if_exists='append', index=False)

        #populate the station data
        query = "SELECT NAME,TYPE,TABLENAME FROM STATIONS_OLD"
        df_stations_table = pd.read_sql_query(query,conn)
        for index in df_stations_table.index:
            station_name = df_stations_table.loc[index]["NAME"]       
            station_id = df_stations[df_stations["NAME"] == station_name].iloc[0]["ID"]
            data_type = df_stations_table.loc[index]["TYPE"]
            table_name = df_stations_table.loc[index]["TABLENAME"]

            print(f"{station_name} - {data_type} - {table_name}")

            if data_type.upper() not in HydroClimateDatabase.station_types:
                continue

            query = "SELECT Date, VALUE FROM " + table_name
            df_data = pd.read_sql_query(query,conn)
            df_data["STATION"] = station_id
            df_data.to_sql(f"STATION_DATA_{HydroClimateDatabase.station_types[data_type.upper()]}",con = conn, if_exists='append',index=False)

            #delete old data table
            HydroClimateDatabase.run_sql(conn, "DROP TABLE " + table_name)
                    
        HydroClimateDatabase.run_sql(conn, "DROP TABLE stations_old")

    @staticmethod
    def migrate(hydroclimate_database_file:str):
        """Migrate the hydroclimate data from old format where the data was stored in separated tables to new format."""

        conn = sqlite3.connect(hydroclimate_database_file)

        with conn:
            if conn is not None:
                HydroClimateDatabase.run_sql(conn, HydroClimateDatabase.sql_change_stations_table_name)
                HydroClimateDatabase.run_sql(conn, HydroClimateDatabase.sql_create_stations_table)

                #create table
                for t in HydroClimateDatabase.station_data_tables:
                    HydroClimateDatabase.run_sql(conn, HydroClimateDatabase.sql_create_station_data_table.replace("STATION_DATA_", t))

                #populate data
                HydroClimateDatabase.populate_stations(conn)

                #create index for station column
                for t in HydroClimateDatabase.station_data_tables:
                    HydroClimateDatabase.run_sql(conn, HydroClimateDatabase.sql_create_station_data_table_index.replace("STATION_DATA_", t))


#endregion


from .bmp_areal import ArealBMP
from ..vector_extension import VectorExtension
from ..database.bmp.bmp_29_manure_feedlot import ManureFeedlot, ManureFeedlotManagement
from ..names import Names
import pandas as pd

class ArealBMPManureFeedlot(ArealBMP):
    """Manure Storage"""
    def __init__(self, bmp_vector, flow_direction_raster, subbasin_raster, reach_raster, dem_raster):
        super().__init__(bmp_vector, flow_direction_raster, subbasin_raster, reach_raster, dem_raster)
        
        self.__manure_feedlot_parameters = None
        self.__manure_feedlot_managements = None

    @property
    def parameters(self)->list:
        if self.__manure_feedlot_parameters is None: 
            self.__manure_feedlot_parameters = []

            feedlot_animal_ids = VectorExtension.get_unique_field_value(self.bmp_vector, Names.field_name_feedlot_animal_id)  
            feedlot_catchbasin_ids = VectorExtension.get_unique_field_value(self.bmp_vector, Names.field_name_feedlot_catch_basin)  
            for id, subbasin in self.subbasins.items():
                self.__manure_feedlot_parameters.append(
                    ManureFeedlot(  id, 
                                    feedlot_animal_ids[id], #animal id must be sepcified
                                    subbasin, 
                                    0 if id not in feedlot_catchbasin_ids else feedlot_catchbasin_ids[id]))

        return self.__manure_feedlot_parameters
    
    @property
    def default_management_df(self)->pd.DataFrame:
        """Default managements populated with the feedlot attributes"""
        if self.__manure_feedlot_managements is None:
            feedlot_adults = VectorExtension.get_unique_field_value(self.bmp_vector, Names.field_name_feedlot_adult)
            feedlot_non_adults = VectorExtension.get_unique_field_value(self.bmp_vector, Names.field_name_feedlot_non_adult)
            feedlot_storage_ids = VectorExtension.get_unique_field_value(self.bmp_vector, Names.field_name_feedlot_storage_ids, str)
            feedlot_storage_ratios = VectorExtension.get_unique_field_value(self.bmp_vector, Names.field_name_feedlot_storage_ratios, str)
            managements = [ManureFeedlotManagement(id, 
                                                   0 if id not in feedlot_adults else feedlot_adults[id],
                                                   0 if id not in feedlot_non_adults else feedlot_non_adults[id],
                                                   "" if id not in feedlot_storage_ids else feedlot_storage_ids[id],
                                                   "" if id not in feedlot_storage_ratios else feedlot_storage_ratios[id],
                                                   ) for id in self.ids]
            
            self.__manure_feedlot_managements = pd.DataFrame([vars(rb) for rb in managements])

        return self.__manure_feedlot_managements


              
       



class Names:
    """
    Default name for input, temp, and output files
    """
    
    bmp_table_name_bmp_index = "bmp_index"
    bmp_table_name_scenarios = "bmp_scenarios"

    #reach bmps
    bmp_table_name_reach_parameter = "reach_parameter"
    bmp_table_name_reach_bmp = "reach_bmp"
    bmp_table_name_point_source = "reach_bmp_point_source"
    bmp_table_name_flow_diversion = "reach_bmp_flow_diversion"
    bmp_table_name_reservoir = 'reach_bmp_reservoir'
    bmp_table_name_wetland = "reach_bmp_wetland"
    bmp_table_name_manure_catch_basin = "reach_bmp_manure_catch_basin"
    bmp_table_name_grass_waterway = 'reach_bmp_grass_waterway'    
    bmp_table_name_managed_access_including_fencing = "reach_bmp_access_management"
    bmp_table_name_water_use = "reach_bmp_water_use"

    #structure bmps
    bmp_table_name_dugout = "structure_bmp_dugout"    
    bmp_table_name_riparian_buffer = "structure_bmp_riparian_buffer"    
    bmp_table_name_filter_strip = "structure_bmp_filter_strip"    
    bmp_table_name_tile_drain = "structure_bmp_tile_drain"
    bmp_table_name_wascob = "structure_bmp_wascob"

    #areal non-structure
    bmp_table_name_crop_management = "non_structure_bmp_crop_management"
    bmp_table_name_crop_parameter = "non_structure_bmp_crop_parameter"
    bmp_table_name_crop_remove_parameter = "non_structure_bmp_crop_remove_parameter"
    bmp_table_name_tillage_management = "non_structure_bmp_tillage_management"
    bmp_table_name_tillage_parameter = "non_structure_bmp_tillage_parameter"
    bmp_table_name_fertilizer_management = "non_structure_bmp_fertilizer_management"
    bmp_table_name_fertilizer_parameter = "fertilizer_parameter"
    bmp_table_name_grazing_management = "GRAMG_management"
    bmp_table_name_livestock_parameter = "livestock_parameter"
    bmp_table_name_irrigation_management = "non_structure_bmp_irrigation_management"
    bmp_table_name_irrigation_parameter = "non_structure_bmp_irrigation_parameter"
    bmp_table_name_manure_storage_parameter = "non_structure_bmp_manure_storage_parameter"
    bmp_table_name_manure_storage_management = "non_structure_bmp_manure_storage_management"    
    bmp_table_name_manure_feed_lot_parameter = "non_structure_bmp_manure_feed_lot_parameter"
    bmp_table_name_manure_feed_lot_management = "non_structure_bmp_manure_feed_lot_management"
    bmp_table_name_wintering_site_parameter = "non_structure_bmp_wintering_site_parameter"
    bmp_table_name_wintering_site_management = "non_structure_bmp_wintering_site_management"
    bmp_table_name_pasture_crop_management = "non_structure_bmp_pasture_crop_management"
    bmp_table_name_pasture_fertilizer_management = "non_structure_bmp_pasture_fertilizer_management"
    bmp_table_name_pasture_tillage_management = "non_structure_bmp_pasture_tillage_management"    
    bmp_table_name_pasture_grazing_management = "non_structure_bmp_pasture_grazing_management"
    bmp_table_name_pasture_grazing_parameter = "non_structure_bmp_pasture_grazing_parametert"
    
    #manure appication adjustment
    bmp_table_name_manure_incorporation_within_48h_management = "manure_adjustment_incorporation_within_48h_management"
    bmp_table_name_manure_application_setback_management = "manure_adjustment_application_setback_management"
    bmp_table_name_manure_no_application_on_snow_management = "manure_adjustment_no_application_on_snow_management"
    bmp_table_name_manure_spring_application_rather_than_fall_application_management = "manure_adjustment_spring_application_rather_than_fall_application_management"
    bmp_table_name_manure_application_based_on_soil_nitrogen_limit_management = "manure_adjustment_application_based_on_soil_nitrogen_limit_management"
    bmp_table_name_manure_application_based_on_soil_phosphorous_limit_management = "manure_adjustment_application_based_on_soil_phosphorous_limit_management"

    #other tables
    bmp_table_name_ls_parameter = "LS_parameter"

    bmp_table_name_farm_info = "farm_info"
    bmp_table_name_field_info = "field_info"
    bmp_table_name_subarea_info = "management_subarea_info"
    bmp_table_name_field_farm = "field_farm"
    bmp_table_name_farm_subbasin = "farm_subbasin"
    bmp_table_name_field_subbasin = "field_subbasin"
    bmp_table_name_subbasin_info = "subbasin_info"
    bmp_table_name_subbasin_multiplier = "subbasin_multiplier"

    #
    bmp_table_name_offsite_watering = "offsite_watering"

    #----------------------------------------------------------------------------

    field_name_id = "id"
    field_name_subbasin = "subbasin"
    field_name_contibution_area_ha = "con_area"
    field_name_raster_value = "VALUE"
    field_name_area = "AREA"

    #the feedlot column in manure storage layer
    field_name_feedlot = "feedlot"

    #feedlot shapefile
    field_name_feedlot_catch_basin = "cb"
    field_name_feedlot_animal_id = "ani_id"
    field_name_feedlot_adult = "adult"
    field_name_feedlot_non_adult = "nonadult"
    field_name_feedlot_storage_ids = "stoid"
    field_name_feedlot_storage_ratios = "storatio"
    fields_feedlot = [
           field_name_feedlot_catch_basin,
            field_name_feedlot_animal_id,
            field_name_feedlot_adult,
            field_name_feedlot_non_adult,
            field_name_feedlot_storage_ids,
            field_name_feedlot_storage_ratios 
    ]


    #-------------------------------------------------------------------------------

    #default raster extension
    raster_extension = ".tif"
    shapefile_extension = ".shp"
    csv_extension = ".csv"
    sqlite_extension = ".db3"
    lookup_extension = ".csv"

    #parameter h5
    parameteH5Name = "parameter.h5"

    #database
    hydroclimateDatabasename = "hydroclimate" + sqlite_extension
    bmpDatabaseName = "bmp" + sqlite_extension
    parameterDatabaseName = "parameter" + sqlite_extension

    # Lookup
    soilLookupName = "soilLookup" + lookup_extension
    landuseLookupName = "landuseLookup" + lookup_extension

    # DEM
    demName = "dem" + raster_extension
    demClippedName = "demClipped" + raster_extension
    demBurnedName = "demClippedBurned" + raster_extension
    demFilledName = "demClippedBurnedFilled" + raster_extension

    # Mask
    boundaryShpName = "boundary" + shapefile_extension
    maskRasName = "mask_before_subbasin_refinement" + raster_extension
    maskRefindedWithSubbasinRasName = "mask" + raster_extension #don't change the name as it will be the base for parameter.h5

    # Flow Direction and Accumulation
    flowDirD8NoChangeName = "flow_dir_no_change" + raster_extension
    flowDirD8Name = "flow_dir" + raster_extension
    flowAccName = "flow_acc" + raster_extension

    # Stream Network
    streamNetworUserShpName = "streamNetworkUser" + shapefile_extension   
    streamNetworUserRasName = streamNetworUserShpName.replace(shapefile_extension, raster_extension)
    
    # streamname = "stream" + shapefile_extension
    # mainStreamRasName = "mainStream" + raster_extension   
    streamMainRasName = "stream_main" + raster_extension
    streamMainShpName = "stream_main" + shapefile_extension 
    streamNetworkRasName = "stream_network" + raster_extension
    streamNetworkShpName = "stream_network" + shapefile_extension
    streamPourPointShpName = "stream_pour_point" + shapefile_extension
    streamPourPointRasName = "stream_pour_point" + raster_extension
    streamOrderRasName = "stream_order" + raster_extension

    # Reach
    reachRasName = "reach" + raster_extension 
    reachShpName = "reach" + shapefile_extension 
    reachParameterCsvName = bmp_table_name_reach_parameter + csv_extension
    reachParameterTxtName = "reachParameter.txt"
    reachDepthName = "reach_depth" + raster_extension
    reachWidthName = "reach_width" + raster_extension

    # Slope
    slopeDegName = "slopeDeg" + raster_extension
    slopeRadiusName = "slopeRadius" + raster_extension
    slopePercentName = "Slope" + raster_extension

    # parameters
    PRCName = "potentialRunoffCoefficient" + raster_extension
    DSCName = "depressionStorageCapacity" + raster_extension
    cn2Name = "cn2" + raster_extension
    PRCAccAvgName = "PRC_Acc_Avg" + raster_extension
    DSCAccAvgName = "DSC_Acc_Avg" + raster_extension
    Cn2AccAvgName = "CN2_Acc_Avg" + raster_extension
    flowLengthName = "flow_length" + raster_extension
    usleKName = "USLE_K" + raster_extension
    uslePName = "USLE_P" + raster_extension
    usleCName = "USLE_C" + raster_extension
    densityName = "density" + raster_extension
    sandName = "sand" + raster_extension
    clayName = "clay" + raster_extension    
    wetnessIndexName = "wetness_index" + raster_extension
    depressionName = "depression" + raster_extension
    residualName = "residual" + raster_extension
    residualIntMaxName = "residualIntMax" + raster_extension
    residualIntMinName = "residualIntMin" + raster_extension
    runoffCoeffName = "runoffCoeff" + raster_extension
    porosityName = "porosity" + raster_extension
    fieldCapName = "field_capacity" + raster_extension
    poreIndexName = "pore_index" + raster_extension
    wiltingPointName = "wilting_point" + raster_extension

    #fixed name, don't change
    moistureInitialName = "Moist_in" + raster_extension 

    manningName = "manning" + raster_extension
    velocityName = "velocity" + raster_extension

    landuseSoilName = "landuseSoil" + raster_extension
    landuseSoil1Name = "landuseSoil1" + raster_extension
    streamWaterName = "streamWater" + raster_extension
    weightName = "weight" + raster_extension
    travelTimeHillT0Name = "travelTimeHillT0" + raster_extension
    travelTimeStreamT0Name = "travelTimeStreamT0" + raster_extension
    travelTimeHillDeltaName = "travelTimeHillDelta" + raster_extension
    travelTimeStreamDeltaName = "travelTimeStreamDelta" + raster_extension    
    slopeLengthName = "slopeLength" + raster_extension
    rootDepthName = "rootDepth" + raster_extension
    coverFractionName = "coverFraction" + raster_extension
    distance2StreamName = "dist_stream" + raster_extension

    windDirSlopeName = "windDirSlope" + raster_extension
    windDirCurvaName = "windDirCurva" + raster_extension
    conductivityName = "conductivity" + raster_extension    

    # Subbasin
    subbasinRasName = "subbasin" + raster_extension
    subbasinShpName = "subbasin" + shapefile_extension

    # Soil 
    soilName = "soil" + raster_extension
    soilMappedName = "soilMapped" + raster_extension

    # Landuse
    landuseName = "landuse" + raster_extension
    landuseMappedName = "landuseMapped" + raster_extension

    # Farm
    farmShpName = "farm" + shapefile_extension
    farmRasName = farmShpName.replace(shapefile_extension, raster_extension)
    farmWithOnlyAgricultureRasName = "farmWithOnlyAgriCulture" + raster_extension    

    # Field
    fieldShpName = "field" + shapefile_extension
    fieldRasName = fieldShpName.replace(shapefile_extension, raster_extension)
    fieldClippedShpName = "fieldClipped" + shapefile_extension
    fiedWithOnlyAgricultureRasName = "fiedWithOnlyAgriCulture" + raster_extension

    # management unit
    subareaShpName = "subarea" + shapefile_extension
    subareaRasName = subareaShpName.replace(shapefile_extension, raster_extension)

    #



    # Outlets
    insertOutletShpName = "insertOutlet" + shapefile_extension
    insertOutletTempShpName = "insertOutletTemp" + shapefile_extension
    insertOutletRasName = "insertOutlet" + raster_extension
    mainOutletsName = "mainOutlets" + shapefile_extension
    outletName = "outlet" + shapefile_extension
    watershedOutletsName = "watershedOutlets" + raster_extension
    streamOutletsName = "streamOutlets" + raster_extension
    streamOutletsWithInsertsName = "streamOutletsWithInserts" + raster_extension
    insertOutletsName = "insertOutlets"
    combinedOutletsName = "combinedOutlets"
   
    #point source
    pointSourceShpName = "point_source" + shapefile_extension
    pointSourceRastername = pointSourceShpName.replace(shapefile_extension, raster_extension)

    #flow diversion
    flowDiversionShpName = "flow_diversion" + shapefile_extension
    flowDiversionRastername = flowDiversionShpName.replace(shapefile_extension, raster_extension)


    # Reservoir
    reservoirShpName = "reservoir" + shapefile_extension
    reservoirRasterName = reservoirShpName.replace(shapefile_extension, raster_extension)

    # Catch basin
    catchbasinShpName = "catchbasin" + shapefile_extension
    catchbasinRasName = catchbasinShpName.replace(shapefile_extension, raster_extension)

    #grass waterway
    grassWaterwayShpName = "grass_waterway" + shapefile_extension
    grassWaterwayRasName = grassWaterwayShpName.replace(shapefile_extension, raster_extension)

    #access management
    accessManagementShpName = "access_management" + shapefile_extension
    accessManagementRasName = accessManagementShpName.replace(shapefile_extension, raster_extension)

    #water use
    waterUseShpName = "water_use" + shapefile_extension
    waterUseRasName = waterUseShpName.replace(shapefile_extension, raster_extension)

    # dugout
    dugoutShpName = "dugout" + shapefile_extension
    dugoutRasName = dugoutShpName.replace(shapefile_extension, raster_extension)

    # wascob
    wascobShpName = "wascob" + shapefile_extension
    wascobRasName = wascobShpName.replace(shapefile_extension, raster_extension)

    #Riparian Buffer
    riparianBufferShpName = "riparian_buffer" + shapefile_extension
    riparianBufferRasterName = riparianBufferShpName.replace(shapefile_extension, raster_extension)
    riparianBufferPartRasterName = "riparian_buffer_part" + raster_extension
    riparianBufferDrainageRasterName = "riparian_buffer_drainage" + raster_extension
    riparianBufferParameterCSVName = "riparian_buffer_parameter" + csv_extension

    #filter strip
    filterStripShpName = "filter_strip" + shapefile_extension
    filterStripRasterName = filterStripShpName.replace(shapefile_extension, raster_extension)
    filterStripPartRasterName = "filter_strip_part" + raster_extension
    filterStripDrainageRasterName = "filter_strip_drainage" + raster_extension
    filterStripParameterCSVName = "filter_strip_parameter" + csv_extension

    #Vegetation Filter Strip    
    vegetationFilterStripShpName = "vegetationFilterStrip" + shapefile_extension
    vegetationFilterStripRasterName = vegetationFilterStripShpName.replace(shapefile_extension, raster_extension)
    # vegetationFilterStripPartRasterName = "vegetationFilterStripPart" + raster_extension
    # vegetationFilterStripPartShpName = "vegetationFilterStripPart" + shapefile_extension
    # vegetationFilterStripDrainageRasterName = "vegetationFilterStripDrainage" + raster_extension
    # vegetationFilterStripDrainageShpName = "vegetationFilterStripDrainage" + shapefile_extension

    #tile drain
    tileDrainShpName = "tile_drain" + shapefile_extension
    tileDrainRasName = tileDrainShpName.replace(shapefile_extension, raster_extension)

    # Cattle Feedlot
    feedlotShpName = "feedlot" + shapefile_extension    
    feedlotOutletShpName = "feedlot_outlet" + shapefile_extension
    feedlotBoundaryProcessedRasName = "feedlotBoundaryProcessed" + raster_extension

    # Manure Storage - MSCD
    manureStorageShpName = "manure_storage" + shapefile_extension
    manureStorageRasName = manureStorageShpName.replace(shapefile_extension, raster_extension)

    # offsite wintering - fixed name - don't change
    offsiteWinteringShpName = "grazingoffsitewatering" + shapefile_extension
    offsiteWinteringRasName = "grazingoffsitewatering" + raster_extension

    #structures
    structureCombinedBoundaryShpName = "structureCombinedBoundary" + shapefile_extension
    structureCombinedOutputShpName = "structureCombinedOutlet" + shapefile_extension
    structureCombinedOutputRasName = "structureCombinedOutlet" + raster_extension

    # Wetland
    wetlandShpName = "wetland" + shapefile_extension
    wetlandRasName = "wetland" + raster_extension
    wetlandInactName = "wetlandInact" + raster_extension
    wetlandAlteredName = "wetlandAltered" + raster_extension
    wetlandDrainAlteredName = "wetlandDrainAltered" + raster_extension
    wetlandDrainLostName = "wetlandDrainLost" + raster_extension
    wetlandClassTypeName = "wetlandClassType" + raster_extension
    wetlandDrainConsolidatedName = "wetlandDrainConsolidated" + raster_extension
    wetlandD8Temp1Name = "wetlandD8Temp1" + raster_extension
    wetlandD8Temp2Name = "wetlandD8Temp2" + raster_extension
    wetlandD8FinalName = "wetlandD8Final" + raster_extension
    wetlandFlowAccTemp1Name = "wetlandFlowAccTemp1" + raster_extension
    wetlandFlowAccTemp2Name = "wetlandFlowAccTemp2" + raster_extension
    wetlandFlowAccFinalName = "wetlandFlowAccFinal" + raster_extension
    wetlandStreamNetTemp1Name = "wetlandStreamNetTemp1" + raster_extension
    wetlandStreamNetTemp1ShpName = "wetlandStreamNetTemp1" + shapefile_extension #
    wetlandStreamNetTemp2Name = "wetlandStreamNetTemp2" + raster_extension
    wetlandStreamNetTemp3Name = "wetlandStreamNetTemp3" + raster_extension
    wetlandStreamNetUserName = "wetlandStreamNetUser" + raster_extension
    wetlandOutletsName = "wetlandOutlets" + raster_extension
    wetlandOutletsUserRasName = "wetlandOutletsUser" + raster_extension    
    wetlandOutletsUserShpName = "wetlandOutletsUser" + shapefile_extension
    wetlandInletRipaName = "wetlandInletRipa" + raster_extension
    wetlandEndNodesName = "wetlandEndNodes" + raster_extension
    wetlandStreamLinkName = "wetlandStreamLink" + raster_extension
    wetlandStreamLinkShpName = "wetlandStreamLink" + shapefile_extension
    wetlandIsoMOutName = "wetlandIsoMOut" + raster_extension
    wetlandIso1OutName = "wetlandIso1Out" + raster_extension
    wetlandIsolateName = "wetlandIsolate" + raster_extension 
    wetlandIsolateShpName = "wetlandIsolate" + shapefile_extension 
    wetlandRipaOutShpName = "wetlandRipaOut" + shapefile_extension 
    wetlandRipaOutName = "wetlandRipaOut" + raster_extension
    wetlandMdfyName = "wetlandMdfy" + raster_extension
    wetlandMdfyShpName = "wetlandMdfy" + shapefile_extension
    wetlandMdfyDissolveShpName = "wetlandMdfyDissolve" + shapefile_extension
    wetlandMdfyDissolvePointsShpName = "wetlandMdfyDissolvePoints" + shapefile_extension
    wetlandMdfyNewIDName = "wetlandMdfyNewID" + raster_extension
    wetlandMdfyNewIDShpName = "wetlandMdfyNewID" + shapefile_extension
    wetlandMdfyNewIDDissolveShpName = "wetlandMdfyNewIDDissolve" + shapefile_extension
    wetlandMdfyNewIDDissolvePointsShpName = "wetlandMdfyNewIDDissolvePoints" + shapefile_extension
    wetlandSubbasinName = "wetlandSubbasin" + raster_extension
    wetlandSubbasinNoWetName = "wetlandSubbasinNoWet" + raster_extension
    wetlandSubbasinWithWetName = "wetlandSubbasinWithWet" + raster_extension
    wetlandSubbasinNoWetShpName = "wetlandSubbasinNoWet" + shapefile_extension
    wetlandSubbasinWithWetShpName = "wetlandSubbasinWithWet" + shapefile_extension
    wetlandInfoName = "WetlandInfo" + ".csv"
    wetlandExtentName = "wetlandExtent" + raster_extension
    wetlandFlowDirMdfyName = "wetlandFlowDirMdfy" + raster_extension
    wetlandUpstreamName = "wetlandUpstream" + raster_extension

    #Wintering Site    
    winteringSiteShpName = "winteringSite" + shapefile_extension
    winteringSiteRasterName = winteringSiteShpName.replace(shapefile_extension,raster_extension)

    #Pasture
    pastureLandRasName = "PasCropland" + raster_extension
    # pastureLandShpName = "PasCropland" + shapefile_extension
    # pastureLandH5RasterName = "PasCropland_H5" + raster_extension

    #Pasture grazing
    pastureGrazingRasterName = "pastureGrazing" + raster_extension
    pastureGrazingShpName = "pastureGrazing" + shapefile_extension

    #Manure Setback
    manure48hShpName = "manure_within_48h" + shapefile_extension
    manure48hRasName = manure48hShpName.replace(shapefile_extension, raster_extension)
    manureSetbackShpName = "manure_setback" + shapefile_extension
    manureSetbackRasName = manureSetbackShpName.replace(shapefile_extension, raster_extension)
    manureNoOnSnowShpName = "manure_no_on_snow" + shapefile_extension
    manureNoOnSnowRasName = manureNoOnSnowShpName.replace(shapefile_extension, raster_extension)
    manureSpringShpName = "manure_spring" + shapefile_extension
    manureSpringRasName = manureSpringShpName.replace(shapefile_extension, raster_extension)
    manureNLimitShpName = "manure_n_limit" + shapefile_extension
    manureNLimitRasname = manureNLimitShpName.replace(shapefile_extension, raster_extension)
    manurePLimitShpName = "manure_p_limit" + shapefile_extension
    manurePLimitRasname = manurePLimitShpName.replace(shapefile_extension, raster_extension)
  

    # streamNetworkNoFieldName = "streamNetworkNoField" + raster_extension
    # averageDistanceToStreamName = "averageDistanceToStream" + raster_extension
    # averageSlopeToStreamName = "averageSlopeToStream" + raster_extension

    # grWaterSourceRasterName = "grWaterSource" + raster_extension
    # grAccessMgtRasterName = "grAccessMgt" + raster_extension
    # grazingRedistributionRasterName = "GrazingRedistribution" + raster_extension
    # grazingOffsiteWateringRasterName = "GrazingOffsiteWatering" + raster_extension

    #Marginal Crop Land
    marCroplandRasName = "MarCropland" + raster_extension
    # marCroplandShpName = "MarCropland" + shapefile_extension
    # marCroplandH5RasterName = "MarCropland_H5" + raster_extension

    #standard names
    config_item_standard_name_lookup = {
        #watershed
        "dem_raster": demName,
        "soil_raster": soilName,
        "landuse_raster": landuseName,
        "stream_shapefile": streamNetworUserShpName,
        "boundary_shapefile": boundaryShpName,
        "farm_shapefile": farmShpName,
        "field_shapefile": fieldShpName,
        "outlet_shapefile": outletName,

        #lookup tables
        "soil_lookup":soilLookupName,
        "landuse_lookup":landuseLookupName,

        #db3 database
        "hydroclimate":hydroclimateDatabasename,

        #reach bmp
        "point_source_shapefile": pointSourceShpName,
        "flow_diversion_shapefile": flowDiversionShpName,
        "reservoir_shapefile": reservoirShpName,
        "wetland_boundary_shapefile": wetlandShpName,
        "wetland_outlet_shapefile": wetlandOutletsUserShpName,
        "manure_catch_basin_shapefile" : catchbasinShpName,
        "grass_waterway_shapefile": grassWaterwayShpName,
        "access_management_shapefile":accessManagementShpName,
        "water_use_shapefile": waterUseShpName,

        #structure_bmp
        "dugout_boundary_shapefile":dugoutShpName,
        "wascob_boundary_shapefile":wascobShpName,
        "riparian_buffer_shapefile":riparianBufferRasterName,
        "filter_strip_shapefile":filterStripRasterName,
        "tile_drain_shapefile":tileDrainShpName,

        #areal non-structure bmp
        "manure_feedlot_boundary_shapefile": feedlotShpName,
        "manure_feedlot_outlet_shapefile": feedlotOutletShpName,
        "manure_storage_boundary_shapefile": manureStorageShpName,
        "offsite_watering_shapefile": offsiteWinteringShpName
    }

    @staticmethod
    def get_standard_file_name(item_name:str)->str:
        if item_name in Names.config_item_standard_name_lookup:
            return Names.config_item_standard_name_lookup[item_name]
        
        raise ValueError(f"{item_name} is not a valide name for standard file name.")
    
    @staticmethod
    def remove_extension_from_raster_file(raster_file_name:str)->str:
        return raster_file_name.replace(Names.raster_extension, "")
    

    





from .bmp_15_fertilizer_management import FertlizerManagement

class PastureFertilizerManagement(FertlizerManagement):
    """Distribution Table for BMP: Pasture fertilizer management (35)"""
    def __init__(self, FerMon, FerDay, FerType, FerRate):
        super().__init__(FerMon, FerDay, FerType, FerRate)


from .bmp_reach import ReachBMP
from whitebox_workflows import Vector, Raster
import pandas as pd

class ReachBMPWetland(ReachBMP):
    def __init__(self, wetland_vector:Vector, subbasin_raster:Raster):
        super().__init__(wetland_vector, subbasin_raster)




# Calculate Watershed Parameters
# 1. Waterhsed.buildWatershedParametersMaps
#     a. Generate mask with subbasin,Landuse and Soil. And copy it as USLEP raster. 
#     b. Calculate parameters with Customized Plugins - Done need some examples
#         1) StreamWidth
#         2) getReachDepthRaster
#         3) ReclassLookup for fc1 and manning
#         4) Velocity
#         5) WetlandIndex
#         6) SoilMoistureInitial
#     c. Watershed.BuildReachParTable - Junzhi Students
#     d. buildDSCRaster
#     e. buildDSCAccumulateAvgMap
#     f. buildCN2Raster
#     g. buildCN2AccumulateAvgMap
#     h. getFieldIDnAreas
#     i. getFarmIDnAreas
#     j. getSubbasinIDnAreas
#     k. createSpatialTables
#     l. buildBMPParMgtTables
#     m. Build livestock parameter table
#     n. Build manure_and_nutrient_parameter table
#     o. Build LS_parameter table
#     p. Build BMP index table
#     q. Build subbasin multiplier table    r. 
# 2. Watersehd.createManureSetbackMap
#     a.ExtractStreams
#     b.ExcludeStreamInField
#     c.AverageSlopeToStream
#     d.createManureSetbackMap

from whitebox_workflows import Raster
from names import inputs, outputs

class watershed:
    def __init__(self, input_folder:str, output_folder:str) -> None:
        self.outputs = outputs(output_folder, input_folder)

    def lookupValues(self, inputRaster:Raster, lookup:dict) -> Raster:
        """
        Replace the raster values based on lookup table. The application could be

        1. generate field capacity raster based on soil lookup table
        2. generate manning raster based on landuse lookup table
        3. 
        
        Parameters:
        inputRaster         : the input raster
        mask                : mask raster
        lookup              : lookup table assuming the input value is the key and output value is the value   
        """

        pass

    def getSlopeRadius(self, slopeDeg:Raster)->Raster:
        return (slopeDeg * math.pi / 180).tan()

    def calculateReachVelocity(self, manning:Raster, slopeRadius:Raster, streamDepth:Raster, mask:Raster, min:float = 0.005, max: float = 3) -> Raster:
        """
        Calculate full reach velocity using Manning equation

        Parameters:
        manning             : Manning raster
        slopeDeg            : Slope raster in degree. Could use radius instead
        streamDepth         : Stream depth raster
        mask                : mask raster
        min                 : min velocity
        max                 : max velocity

        """
        velocity = slopeRadius.sqrt() * streamDepth**0.6667 / manning
        return velocity.min(max).max(min)

    def calculateWetnessIndex(self, flowAcc:Raster, slopeRadius:Raster, mask:Raster)->Raster:
        """
        Calculate wetness index using flow accumulation and slope
        """

        return (flowAcc * getRasterCellAreaM2(mask)).log() / slopeRadius.tan()

    def calculateIntialSoilMoisture(self, wetnessIndex:Raster, fieldCapacity:Raster, mask:Raster, minSaturation:float=0.05, maxSaturation:float=1)->Raster:
        """
        Calculate the initial soil moisture using linear interpolation
        """

        minWetnessIndex = wetnessIndex.configs.minimum
        maxWetnessIndex = wetnessIndex.configs.maximum * 0.8

        wti = wetnessIndex.max(maxWetnessIndex)
        ratio = (wti - minWetnessIndex) * (maxSaturation - minSaturation) / (maxWetnessIndex - minWetnessIndex) + minSaturation
        return ratio * fieldCapacity

    def buildReachParameterTable():
        pass

    def calculatePotentialRunoffCoefficient(self, lope:Raster, mask:Raster, landuse:Raster, soil:Raster, reach:Raster, flowDirectionD8:Raster, reachWidth:dict)->Raster:
        
        #calculate reach water surface fraction
        #need to find a quick way to do lookup
        reachWaterSurfaceFraction = this.getReachWidth(reachID) / mask.configs.resolution_x    
        reachWaterSurfaceFraction = flowDirectionD8.con("(value == 1) || (value ==4) || (value == 8) || (value ==16)", 1.41421356 * reachWaterSurfaceFraction - 0.5 * reachWaterSurfaceFraction * reachWaterSurfaceFraction, reachWaterSurfaceFraction)

    def getPotentialRunoffCoefficient(landuseId:int, soilId:int, slope:float, reachFraction:float, landuseParameterLookup:pd.DataFrame, soilParameterLookup:pd.DataFrame):
        """
        
        Soid Id -> Soil Texture (1-12
        LanUse Id -> PRC_ST1, ..., PRC_ST12 where ST is for soil texture
        """






from sqlalchemy import create_engine
import pandas as pd
import os
import logging
logger = logging.getLogger(__name__)


class DatabaseBase:
    def __init__(self, database_file):
        # if not os.path.exists(database_file):
        #     raise ValueError(f"{database_file} doesn't exist.")
        
        self.database_file = database_file        
        self.engine = create_engine(f'sqlite:///{self.database_file}')  

    def read_single_value(self, table_name:str, filter_column_name:str, filter_value:str, value_column_name:str):
        """read a single value from a table with filter by another column"""
        if not self.check_table_exist(table_name):
            raise ValueError(f"Table: {table_name} doesn't exist in {self.database_file}.")

        df = pd.read_sql(f'select {value_column_name} from {table_name} where {filter_column_name} = "{filter_value}"', self.engine)
        if len(df) > 0:
            return df[df.columns[-1]].to_list()[0]

    def read_distinct_list(self, table_name:str, column_name:str)->pd.DataFrame:
        """read distinct list of given column in given table"""
        return pd.read_sql(f"select distinct {column_name} from {table_name}", self.engine)

    def read_table(self, table_name:str)->pd.DataFrame:
        """read the whole table and return dataframe"""
        if not self.check_table_exist(table_name):
            return None
        return pd.read_sql(f"select * from {table_name}", self.engine)

    def save_table(self, table_name:str, table_df:pd.DataFrame, dtype:list = None, index = False):
        table_df.to_sql(table_name, con = self.engine, if_exists='replace',index=index, dtype=dtype)

    def check_table_exist(self, table_name:str)->bool:
        return len(pd.read_sql(f"SELECT tbl_name FROM sqlite_master where type='table' and tbl_name ='{table_name}'",self.engine)) > 0 

    def populate_defaults(self, table_name:str):
        """
        populate table from csv file, used to load the default tables, assuming the table name is same as the csv file name
        """
        
        # if self.check_table_exist(table_name):
        #     logger.info(f"Table {table_name} already exist, skip")
        #     return

        #try to find the corresponding csv file in the default folder
        csv_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "defaults", f"{table_name}.csv")

        #raise error if the csv file doesn't exist
        if not os.path.exists(csv_file):
            raise ValueError(f"{csv_file} doesn't exist.")
        
        #load csv file and save to the database. It will replace the existing table.
        try:
            df = pd.read_csv(csv_file)
            df.to_sql(table_name, con = self.engine, if_exists='replace', index = False)
        except:
            raise ValueError(f"{csv_file} couldn't be imported to table {table_name} in database: {self.database_file}")
        


from sqlalchemy import Column, Integer, Float
from .bmp_table import BMPTable

class FieldSubbasin(BMPTable):
    __tablename__ = 'field_subbasin'
    ID = Column(Integer, primary_key=True)
    Field = Column(Integer)
    Subbasin = Column(Integer)
    Area_Ha = Column(Float)
    ToField = Column(Float)
    ToSubbasin = Column(Float)


from whitebox_workflows import Raster

class Interpolation:
    def __init__(self, weight_file:str) -> None:
        self.weight_file = weight_file

    def write_weight_file(self, mask_raster:Raster, station_coordinates:list):
        """write weight file"""
        pass



from .bmp_12_crop_management import CropManagement

class PastureCropManagement(CropManagement):
    def __init__(self, CropCode, PlantingMon, PlantingDay, HarvestMon, HarvestDay):
        super().__init__(CropCode, PlantingMon, PlantingDay, HarvestMon, HarvestDay)





from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine, select
from ..database_base import DatabaseBase
from .landuse_lookup import LanduseLookup
from .soil_lookup import SoilLookup
import logging
logger = logging.getLogger(__name__)

class ParameterDatabase(DatabaseBase):
    """Access to parameter database."""

    default_tables = ["Climate", 
                      "Discharge", 
                      "Interception",
                      "NutrientCycling",
                      "PlantGrowth",
                      "Sediment",
                      "Snow",
                      "SubArea",
                      "WaterBalance",
                      "Wetland",
                      "LanduseLookup",
                      "SoilLookup", 
                      "LanduseSoilLoopUpCalibration",
                      "Lapse_rate", 
                      "LS_parameter",
                      "tile_drain"]

    def __init__(self, database_file):
        super().__init__(database_file) 

        self._lookup_tables = {}
        self._soil_lookup = {}
        self._landuse_lookup = {}
        self._algricultural_landuses = []

    def __populate_default_tables(self):
        """populate default parameter tables from csv files"""

        #populate default tables
        logger.info("Trying to load default tables to parameter database ...")
        for table in ParameterDatabase.default_tables:
            logger.info(table)
            self.populate_defaults(table)

#region Soil and Landuse Parameters


    @property
    def soil_lookup(self)->dict:
        if len(self._soil_lookup) == 0:
            self.__populate_default_tables()

            Session = sessionmaker(bind=self.engine)
            with Session() as session:
                select_stmt = select(SoilLookup)
                for row in session.scalars(select_stmt):
                    self._soil_lookup[row.SOILCODE] = row

        return self._soil_lookup
    
    @property
    def landuse_lookup(self)->dict:
        if len(self._landuse_lookup) == 0:
            self.__populate_default_tables()

            Session = sessionmaker(bind=self.engine)
            with Session() as session:
                select_stmt = select(LanduseLookup)
                for row in session.scalars(select_stmt):
                    self._landuse_lookup[row.LANDUSE_ID] = row
                    if row.is_agricultrual:
                        self._algricultural_landuses.append(row.LANDUSE_ID)

        return self._landuse_lookup

    def get_parameter_lookup(self, parameter_name, parameter_type):
        """return lookup array to be used in reclass funciton"""

        if parameter_name in self._lookup_tables:
            return self._lookup_tables[parameter_name]
        
        lookups = []
        if parameter_type == "soil":
            values = self.soil_lookup.items()
        elif parameter_type == "landuse":
            values = self.landuse_lookup.items()
        else:
            raise ValueError(f"Invalide parameter type {parameter_type}. Please use either soil or landuse.")
        for key, value in values:
            lookups.append([key, getattr(value, parameter_name)])
        
        self._lookup_tables[parameter_name] = lookups

        return lookups

    def get_soil_texture(self, soil:str)->int:
        if soil not in self.soil_lookup:
            raise ValueError(f"Soil {soil} doesn't exist in parameter soil lookup table.")
        return int(self.soil_lookup[soil].TEXTURE)

    def get_potential_runoff_coefficient(self, landuse, soil:str)->float:
        """Get potential runoff coefficient for given landuse and soil"""
        return self.landuse_lookup[landuse].getPotentialRunoffCoefficient(self.get_soil_texture(soil))

    def get_depression_storage_capacity(self, landuse, soil)->float:
        """Get depression storage capacity for given landuse and soil"""
        return self.landuse_lookup[landuse].getDepressionStorageCapacity(self.get_soil_texture(soil))

    def get_slope(self, landuse, soil)->float:
        """Get slope for given landuse and soil"""
        return self.landuse_lookup[landuse].getSlope(self.get_soil_texture(soil))

    def get_cn2(self, landuse, soil)->float:
        """Get CN2 for given landuse and soil"""
        return self.landuse_lookup[landuse].getCN2(self.soil_lookup[soil].HG)

    def get_impervious_ratio(self, landuse)->float:        
        """Get impervious ration for given landuse"""
        return self.landuse_lookup[landuse].FIMP
    
#endregion

    
#region Single Parameter

    def __get_single_parameter(self, parameter_table_name:str, parameter_name:str)->float:
        """get parameter value with given name and table"""
        return float(self.read_single_value(parameter_table_name, "Parameter", parameter_name, "Value"))

    @property
    def wetland_co1(self)->float:
        return self.__get_single_parameter("Wetland","Wet_Co1")
    
    @property
    def wetland_ex1(self)->float:
        return self.__get_single_parameter("Wetland","Wet_Ex1")   
    
    @property
    def wetland_co2(self)->float:
        return self.__get_single_parameter("Wetland","Wet_Co2")
    
    @property
    def wetland_cons(self)->float:
        return self.__get_single_parameter("Wetland","Wet_Cons")  
    
#endregion


import os
import shutil

class ConfigFile:
    default_file_name = "config.fig"

    def __init__(self, folder) -> None:
        self.folder = folder

    def write_file(self):
        """Just copy the default file for now"""
        default_config_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "defaults", ConfigFile.default_file_name)
        destination_file = os.path.join(self.folder, ConfigFile.default_file_name)
        shutil.copyfile(default_config_file, destination_file)
        



from sqlalchemy import Column, Integer
from .bmp_table import BMPTable
from ...names import Names

class ManagedAccessIncludingFencingParameter(BMPTable):
    """Parameter Table for BMP: Managed access inluding fencing (40)"""
    __tablename__ = Names.bmp_table_name_managed_access_including_fencing
    """Reach ID"""
    ID = Column(Integer, primary_key = True)
    """Subbasin ID"""
    Subbasin = Column(Integer)
    """Operation starting year"""
    Year = Column(Integer)



import pandas as pd
import os
import numpy as np
import random
from ..bmp import BMP
from ..bmp_type import DefaultScenarioId
from whitebox_workflows import Vector, Raster
from ...raster_extension import RasterExtension
from ...vector_extension import VectorExtension
from ...folder_base import FolderBase

from ...database.bmp.bmp_12_crop_management import CropManagement
from ...database.bmp.bmp_15_fertilizer_management import FertlizerManagement
from ...database.bmp.bmp_14_tillage_management import TillageManagement
from ...database.bmp.bmp_16_grazing_management import GrazingManagement

import logging
logger = logging.getLogger(__name__)

class CropRotation(FolderBase):
    flag_not_proceed = "Not proceed"    
    col_location = "Location"
    col_year = "Year"
    col_original_crop_id = "Original ID"
    col_imwebs_crop_id = "IMWEBs ID"
    col_crop_code = "CropCode"

    #create random date between first and last day
    col_planting_date = ('PlantingMon', 'PlantingDay')
    col_harvest_date = ('HarvestMon', 'HarvestDay')
    col_fertilizer_date = ('FerMon', 'FerDay')
    col_tillage_date = ('TillMon', 'TillDay')
    value_use_planting_date = -1
    value_30_after_harvest = -30

    def __init__(self, 
                 management_unit_vector:Vector,
                 subbasin_raster: Raster,
                 crop_inventory_raster_folder:str, 
                 first_year:int, 
                 last_year:int,
                 scenario:int = DefaultScenarioId):
        super().__init__(crop_inventory_raster_folder)
        self.management_unit_vector = management_unit_vector
        self.subbasin_raster = subbasin_raster
        self.first_year = first_year
        self.last_year = last_year
        self.scenario = scenario

        exist, self.id_field_name = VectorExtension.check_id(self.management_unit_vector)
        if not exist:
            raise ValueError(f"There is no id column in {self.management_unit_vector.file_name}.")

        self.__dominant_crop_df = None
        self.__default_setup_dfs = {}
        self.__crop_management_df = None
        self.__fertilizer_management_df = None
        self.__tillage_management_df = None
        self.__grazing_management_df = None

    def __get_default_file_path(self, file_name)->str:
        file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), file_name)
        if not os.path.exists(file_path):
            raise ValueError(f"{file_path} doesn't exist. Please check the crop rotation folder.")
        
        return file_path

    def __get_default_setup_df(self, file_name:str)->pd.DataFrame:
        if file_name not in self.__default_setup_dfs:
            df = pd.read_csv(self.__get_default_file_path(file_name))  
            if "crop-lookup" in file_name:
                df = df[df[self.col_imwebs_crop_id]!= self.flag_not_proceed]
                df[self.col_imwebs_crop_id] = df[self.col_imwebs_crop_id].astype(int)
            self.__default_setup_dfs[file_name] = df

        return self.__default_setup_dfs[file_name]

    def __get_crop_code_lookup_df(self, year):
        return self.__get_default_setup_df("crop-lookup-before-2010.csv") if year <=2010 else self.__get_default_setup_df("crop-lookup-after-2010.csv")

    @property
    def domimant_crop_df(self)->pd.DataFrame:
        """domminant crop dataframe for each year"""
        if self.__dominant_crop_df is None:
            field_crop_df_list = []
            for year in range(self.first_year, self.last_year + 1):
                file_name = f"{year}.tif"
                logger.info(file_name)
                crop_raster = self.get_raster(file_name)
                if crop_raster is None:
                    raise ValueError(f"Couldn't find {self.get_file_path(file_name)}.")

                #clip the crop raster to the field vector
                #logger.info("clip ...")
                clipped_crop_raster = self.wbe.clip_raster_to_polygon(raster = crop_raster, polygons = self.management_unit_vector)

                #covert the field vector to raster  
                #logger.info("to raster ...")            
                field_raster = self.wbe.vector_polygons_to_raster(
                                                        input=self.management_unit_vector, 
                                                        base_raster = clipped_crop_raster, 
                                                        field_name = self.id_field_name)
                
                #get dominant crop for each field
                #logger.info("get majority count ...")
                field_dominant_crop = RasterExtension.get_majority_count(clipped_crop_raster, field_raster)

                #create dataframe and save the original crop id to Original ID column
                df = pd.DataFrame.from_dict(field_dominant_crop, orient = 'index', columns = [self.col_original_crop_id])
                df.index.name = self.col_location
                df = df.sort_index()
                df = df.reset_index()
                df['ActualYear'] = year
                df['Year'] = year - self.first_year + 1

                #map to imwebs crop id
                #read crop lookup csv file and dicard not proccessed codes
                df_crop_code_lookup = self.__get_crop_code_lookup_df(year)               

                #join original id and imweb code
                df_field_crop_imwebs = df.merge(df_crop_code_lookup, how = 'inner', on = self.col_original_crop_id)
                df_field_crop_imwebs[self.col_crop_code] = df_field_crop_imwebs[self.col_imwebs_crop_id]

                #add to the list
                field_crop_df_list.append(df_field_crop_imwebs)

            self.__dominant_crop_df = pd.concat(field_crop_df_list, ignore_index=True)
            self.__dominant_crop_df ['Scenario'] = self.scenario
        
        return self.__dominant_crop_df

    @property
    def crop_management_df(self)->pd.DataFrame:
        if self.__crop_management_df  is None:
            #crop management
            df_crop_management_lookup = self.__get_default_setup_df("crop-management.csv")
            df_crop = self.domimant_crop_df.merge(df_crop_management_lookup, how = 'inner', on = self.col_crop_code)

            #add planting and harvest month and day column
            for col in self.col_planting_date + self.col_harvest_date:
                df_crop[col] = -1
                
            #add ID column
            df_crop['ID'] = df_crop['Location']

            #populate the columns
            for index in df_crop.index:    
                #planting dates
                CropRotation.populate_random_date(df_crop, index, self.col_planting_date[0], self.col_planting_date[1])
                
                #harvest dates
                CropRotation.populate_random_date(df_crop, index, self.col_harvest_date[0], self.col_harvest_date[1])

            for index in df_crop.index: 
                #change the planting date of crops that are planted in previous year
                CropRotation.move_winter_wheat_planting_day(df_crop, index)

            #Get the year. If previous year flag is 1, then previous year is used.
            df_crop['Year'] = np.where(df_crop['Pervious Year'] == 1, df_crop['Year'] - 1, df_crop['Year'])

            #if the year is zero, then it's winter wheat and we will use year 1 and set plant day to Jan 1st
            df_crop[self.col_planting_date[0]] = np.where(df_crop['Year'] == 0, 1, df_crop[self.col_planting_date[0]])
            df_crop[self.col_planting_date[1]] = np.where(df_crop['Year'] == 0, 1, df_crop[self.col_planting_date[1]])
            df_crop['Year'] = np.where(df_crop['Year'] == 0, 1, df_crop['Year'])
                
            #save crop managment
            self.__crop_management_df = df_crop[['Location','Scenario','ID','ActualYear','CropCode',
                    'Year','PlantingDay','PlantingMon','HarvestDay','HarvestMon',
                    'HarvestEfficiency','HarvestType','HarvestIndexOverride','CNOP',
                    'StoverFraction','IsGrain','PRCOP']]
            
        return self.__crop_management_df

    @property
    def fertilizer_management_df(self)->pd.DataFrame:
        if self.__fertilizer_management_df is None:
            df_fer_management_lookup = self.__get_default_setup_df("fertilizer-management.csv")
            df_fertilizer = self.crop_management_df.merge(df_fer_management_lookup, how = 'inner', on = self.col_crop_code)

            #add the columns
            for col in self.col_fertilizer_date:
                df_fertilizer[col] = -1

            #populate the columns
            for index in df_fertilizer.index:    
                #fertilizer dates
                CropRotation.populate_random_date(df_fertilizer, index, self.col_fertilizer_date[0], self.col_fertilizer_date[1])
                
            #save fertilizer managment
            self.__fertilizer_management_df = df_fertilizer[['Scenario','Location','Year','FerMon','FerDay',
                                                            'FerType','FerRate','FerSurface']]
        return self.__fertilizer_management_df

    @property
    def tillage_management_df(self)->pd.DataFrame:
        if self.__tillage_management_df is None:
               #-------------------------------------------------------------------------------------------------
            #tillage management   
            df_til_management_lookup = self.__get_default_setup_df("tillage-management.csv")
            df_tillage = self.crop_management_df.merge(df_til_management_lookup, how = 'inner', on = self.col_crop_code)
                
            #add the columns
            for col in self.col_tillage_date:
                df_tillage[col] = -1
                    
            #populate the columns
            for index in df_tillage.index:
                #tillage dates
                CropRotation.populate_random_date(df_tillage, index, self.col_tillage_date[0], self.col_tillage_date[1])
                
            #save tillage managment
            self.__tillage_management_df = df_tillage[['Scenario','Location','Year','TillMon','TillDay','TillCode']]

        return self.__tillage_management_df

    @property
    def grazing_management_df(self)->pd.DataFrame:
        if self.__grazing_management_df is None:
            df_grazaing_management_lookup = self.__get_default_setup_df("grazing-management.csv")
            df_grazing = self.domimant_crop_df.merge(df_grazaing_management_lookup, how = 'inner', on = self.col_crop_code)

            #get field to subbasin mapping
            id_subbasin_dict = BMP(self.management_unit_vector, self.subbasin_raster).subbasins
            df_field_subbasin_lookup = pd.DataFrame(list(id_subbasin_dict.items()),columns=["Location","SourceID"])
            df_grazing = df_grazing.merge(df_field_subbasin_lookup, how = 'inner', on = 'Location')

            #save grazing managment
            self.__grazing_management_df = df_grazing[['Scenario','Location','Year','GraMon','GraDay','Days',
            'Ani_ID','Ani_adult','GR_Density','DayFra','Source', 'SourceID','Dugout_ID','Access','Fencing','StreamAniPerc', 'Drinking_time','BankK_Change']]

        return self.__grazing_management_df

#region select random date

    @staticmethod
    def create_random_date(mon_first, day_first, mon_last, day_last):
        """create random planting month and day"""
        first_day = pd.Timestamp(year = 2000, month = mon_first, day = day_first)
        last_day = pd.Timestamp(year = 2000, month = mon_last, day = day_last)
        day_range = last_day - first_day
        ran = random.randint(0, day_range.days)
        ran_day = first_day + pd.Timedelta(f'{ran} days')
        return (ran_day.month, ran_day.day)

    @staticmethod
    def offset_days(mon, day, offset_days):
        d = pd.Timestamp(year = 2000, month = mon, day = day)
        d = d + pd.Timedelta(f'{offset_days} days')
        
        return (d.month, d.day)

    @staticmethod
    def populate_random_date(df, index, col_mon, col_day):
        """populate the random date for given index and mon/day column"""
        mon_first = df.loc[index, f'{col_mon}_First']
        day_first = df.loc[index, f'{col_day}_First']
        mon_last = df.loc[index, f'{col_mon}_Last']
        day_last = df.loc[index, f'{col_day}_Last']
        
        
        #if one of the mon/day is -1, then planting month/day will be used
        if mon_first == CropRotation.value_use_planting_date | day_first == CropRotation.value_use_planting_date | mon_last == CropRotation.value_use_planting_date | day_last == CropRotation.value_use_planting_date: 
            df.loc[index, col_mon] = df.loc[index, CropRotation.col_planting_date[0]]
            df.loc[index, col_day] = df.loc[index, CropRotation.col_planting_date[1]]
        #if one of the mon/day is -30, then use 30 days after harvest date
        elif mon_first == CropRotation.value_30_after_harvest | day_first == CropRotation.value_30_after_harvest| mon_last == CropRotation.value_30_after_harvest | day_last == CropRotation.value_30_after_harvest: 
            offset = CropRotation.offset_days(df.loc[index, CropRotation.col_harvest_date[0]], df.loc[index, CropRotation.col_harvest_date[1]], 30)
            df.loc[index, col_mon] = offset[0]
            df.loc[index, col_day] = offset[1]        
        else:    
            ran = CropRotation.create_random_date(mon_first, day_first, mon_last, day_last)    
            df.loc[index, col_mon] = ran[0]
            df.loc[index, col_day] = ran[1]

    @staticmethod
    def move_winter_wheat_planting_day(df, index):
        """
            make sure the planting day of winter wheat is after the harvest day of previous crop
            if not, set it to the harvest day of previous crop
        """
        #only apply to crop where the previous year flag is 1
        if df.loc[index, 'Pervious Year'] != 1:
            return
        #only apply to crop on the second years and after as the first year will be set to Jan 1st
        year = df.loc[index, 'Year']
        if year == 1:
            return
        #get crop of previous year
        location = df.loc[index, 'Location']
        previous_crop_df = df[(df['Year'] == year - 1) & (df['Location'] == location)]

        #return if previous crop doesn't exist
        if len(previous_crop_df.index) == 0:
            return

        #get harvest date
        harvest_mon = previous_crop_df.loc[previous_crop_df.index[0],CropRotation.col_harvest_date[0]]
        harvest_day = previous_crop_df.loc[previous_crop_df.index[0],CropRotation.col_harvest_date[1]]
        harvest_date = pd.Timestamp(year = 2000, month = harvest_mon, day = harvest_day)

        #get planting date
        planting_mon = df.loc[index,CropRotation.col_planting_date[0]]
        planting_day = df.loc[index,CropRotation.col_planting_date[1]]
        planting_date = pd.Timestamp(year = 2000, month = planting_mon, day = planting_day)

        #print(f"Location = {location}, Year = {year}, Planting = {planting_mon}-{planting_day}, Previous Harvest = {harvest_mon}-{harvest_day}")

        #check planting date
        if planting_date >= harvest_date:
            return

        #chang it to harvest date
        df.loc[index,CropRotation.col_planting_date[0]] = harvest_mon
        df.loc[index,CropRotation.col_planting_date[1]] = harvest_day

        #print(f"Changed, Planting = {df.loc[index,col_planting_date[0]]}-{df.loc[index,col_planting_date[1]]}")


#endregion




from sqlalchemy import Column, Integer, Float
from sqlalchemy import create_engine, select
from .bmp_table import BMPTable
from ...names import Names

class ReachParameter(BMPTable):
    """Reach parameter in BMP database"""
    __tablename__ = Names.bmp_table_name_reach_parameter

    # Define columns
    reach_id = Column(Integer, primary_key=True)
    subbasin_id = Column(Integer)
    outlet_id = Column(Integer)
    length = Column(Float)
    width = Column(Float)
    main_side_slope = Column(Float)
    floodplain_side_slope = Column(Float)
    w_ratio = Column(Float)
    depth = Column(Float)
    slope = Column(Float)
    order = Column(Integer)
    manning = Column(Float)
    velocity = Column(Float)
    t0 = Column(Float)
    d0 = Column(Float)
    max_elevation = Column(Float)
    min_elevation = Column(Float)
    Ave_elevation = Column(Float)
    receive_reach_id = Column(Integer)
    contribution_area = Column(Float)
    erodibility_factor = Column(Float)
    cover_factor = Column(Float)
    bc1 = Column(Float)
    bc2 = Column(Float)
    bc3 = Column(Float)
    bc4 = Column(Float)
    rk1 = Column(Float)
    rk2 = Column(Float)
    rk3 = Column(Float)
    rk4 = Column(Float)
    rs1 = Column(Float)
    rs2 = Column(Float)
    rs3 = Column(Float)
    rs4 = Column(Float)
    rs5 = Column(Float)
    minFlow = Column(Float)
    k_chb = Column(Float)
    k_bank = Column(Float)
    bnk0 = Column(Float)
    chs0 = Column(Float)
    a_bnk = Column(Float)
    b_bnk = Column(Float)
    MSK_X = Column(Float)
    MSK_col = Column(Float)

    def __init__(self):
        self.main_side_slope = 0.5
        self.floodplain_side_slope = 0.2
        self.w_ratio = 5.0
        self.t0 = 0
        self.d0 = 0
        self.erodibility_factor = 0.1
        self.cover_factor = 0.2
        self.bc1 = 0.55
        self.bc2 = 1.1
        self.bc3 = 0.21
        self.bc4 = 0.35
        self.rk1 = 1.71
        self.rk2 = 50
        self.rk3 = 0.36
        self.rk4 = 2
        self.rs1 = 1.0
        self.rs2 = 0.05
        self.rs3 = 0.5
        self.rs4 = 0.05
        self.rs5 = 0.05
        self.minFlow = 0
        self.k_chb = 0
        self.k_bank = 0
        self.bnk0 = 0
        self.chs0 = 0
        self.a_bnk = 0.2
        self.b_bnk = 0.05
        self.MSK_X = 0
        self.MSK_col = 0
        
    @staticmethod
    def get_columns():
        r = ReachParameter()
        columns = []
        for col in r.__table__.columns:
            if hasattr(r, col.name):
                columns.append(col.name)

        return columns


from .interpolation import Interpolation
from .average_uniform import AverageUniform
from .grid_interpolation import GridInterpolation
from .inverse_distance import InverseDistance
from .linear_triangle import LinearTriangle
from .thiessen_polygon import ThiessenPolygon
from whitebox_workflows import Raster
import logging
logger = logging.getLogger(__name__)

def WriteWeightFile(method:str, weight_file:str,mask_raster:Raster, station_coordinates:list):
    #get interpolation object based on method
    interploation = Interpolation(weight_file)
    if method == "average_uniform":
        interploation = AverageUniform(weight_file)
    elif method == "grid_interpolation":
        interploation = GridInterpolation(weight_file)
    elif method == "inverse_distance":
        interploation = InverseDistance(weight_file)
    elif method == "linear_triangle":
        interploation = LinearTriangle(weight_file)
    elif method == "thiessen_polygon":
        interploation = LinearTriangle(weight_file)
    else:
        logger.info(f"Interpolation method {method} is not valide. Use inverst distance method instead.")
        interploation = InverseDistance(weight_file)

    #generate the file
    interploation.write_weight_file(mask_raster, station_coordinates)
    
